/* stl2gts.c - convert STL triangles to GTS surface */

/* setreuid(2) */
#define _XOPEN_SOURCE  500

#include <errno.h>
#include <inttypes.h>
#include <locale.h>
#include <math.h>
#include <regex.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#define G_DISABLE_DEPRECATED  1
#include <gts.h>


#define CHRNUL  '\0'

#define EEXIT  (eExit(__func__,__LINE__))

#define LNBUFSZ  2048

/* ascii stl only uses floats */
#define RE_N  "[-+]?([[:digit:]]*[.]?[[:digit:]]+|[[:digit:]]+[.][[:digit:]]*)"
#define RE_E  "([Ee][-+]?[[:digit:]]+)?"
#define REFLT "(" RE_N RE_E ")"
#define RESPC  "[ \t]"

/* triangle (read) array, 50 bytes.
 * binary STL triangles are four float 3-vectors, plus a short,
 * which is always ignored/discarded (as is the normal vector).
 */
#define TSZ  50
typedef union tXfer tXfer;
union tXfer {
  uint8_t raw[TSZ];
  float   val[12];  /* 0,3,6,9 */
};

typedef struct tVertex tVertex;
struct tVertex {float x, y, z;};
typedef tVertex tTriangle[3];


/* ascii format stuff */
static char const * const ascident = "solid ";
static char const * const ascends  = "endsolid";

enum REIndex {reVertex = 0, reFacet, reFacetEnd, reLoop, reLoopEnd, reEndRE};
typedef enum REIndex REIndex;

static char const * const repatterns [reEndRE] = {
 "^" RESPC "*" "vertex" RESPC "+" REFLT RESPC "+"
         REFLT RESPC "+" REFLT RESPC "*",
 "^" RESPC "*" "facet normal" RESPC "+" REFLT RESPC "+"
         REFLT RESPC "+" REFLT RESPC "*",
 "^" RESPC "*" "endfacet" RESPC "*" "\n$",
 "^" RESPC "*" "outer loop" RESPC "*" "\n$",
 "^" RESPC "*" "endloop" RESPC "*" "\n$"
};
static regex_t rgx[reEndRE];

/* CLI stuff */
static char const * const prgname = "stl2gts";
static char const * const prgversion = "0.1.0";
/* -D    diagnstic stats.
   -h    help.
   -V    version.  (share help message).
 */
static char const * const prgopts = "+DhV";
static char const * const prgopts_help[] = {
  "D\tdisplay diagnostic stats",
  "h\tdisplay this message, and exit",
  "V\tdisplay the program version, and exit",
  NULL
};
static bool diag;

static bool ascfmt;
static uint32_t tcount;

static char const * const fmtunexpected =
        "%s: oops, unexpected (end of) data in line %u.\n";

static GtsSurface *S;


/* -------------------------------------------------------------------------- *
 * oops + GNU mandated interface.
 */

static void eExit(char const * const, int const) __attribute__((noreturn));
static void eExit(char const * const fn, int const ln) {
  int const e = errno;
  fprintf(stderr, "%s: oops in '%s' line %d, error (%d) %s.\n",
          prgname, fn, ln, e, strerror(e));
  exit(EXIT_FAILURE);
}

static void usageHelp(int const) __attribute__((noreturn));
static void usageHelp(int const ec) {
  char const * const fmt = "%s - STL triangles to GTS surface.\n"
          "\n  usage: %s [opts] < thing.stl > thing.gts\n"
          "\navailable options are:\n";
  printf(fmt, prgname, prgname);
  for (size_t i = 0; prgopts_help[i]; ++i)
      printf("  -%s.\n", prgopts_help[i]);
  printf("\n");
  exit(ec);
}

static void usageVersion(void) __attribute__((noreturn));
static void usageVersion(void) {
  char const * const fmt = "%s %s\n"
          "copyright (c) 2020, jr <creature.eternal@gmail.com>.\n"
          "license GPLv3+ <http://gnu.org/licenses/gpl.html>\n"
          "this is free software, you are free to .. whatever.\n"
          "there is no warranty.\n";
  printf(fmt, prgname, prgversion);
  exit(EXIT_SUCCESS);
}

/* -------------------------------------------------------------------------- *
 * process data.
 */

static bool getLine(char * const line) {
  if (!fgets(line, LNBUFSZ, stdin)) {
    if (feof(stdin))
      return false;
    EEXIT;
  }
  return true;
}

static bool addTriangle(tTriangle const raw) {
  GtsVertex *v1 =
          gts_vertex_new(gts_vertex_class(), raw[0].x, raw[0].y, raw[0].z);
  GtsVertex *v2 =
          gts_vertex_new(gts_vertex_class(), raw[1].x, raw[1].y, raw[1].z);
  GtsVertex *v3 =
          gts_vertex_new(gts_vertex_class(), raw[2].x, raw[2].y, raw[2].z);
  GtsEdge *e1 = gts_edge_new(gts_edge_class(), v1, v2);
  GtsEdge *e2 = gts_edge_new(gts_edge_class(), v2, v3);
  GtsEdge *e3 = gts_edge_new(gts_edge_class(), v3, v1);
  GtsFace *t = gts_face_new(gts_face_class(), e1, e2, e3);
  gts_surface_add_face(S, t);
  tcount += 1;
  return true;
}

static bool procOloop(uint32_t *lnum) {
  char line[LNBUFSZ];
  memset(line, 0, LNBUFSZ);
  *lnum += 1;
  if (!getLine(line)) {
    fprintf(stderr, fmtunexpected, prgname, *lnum);
    return false;
  }
  else if (regexec(&rgx[reLoop], line, 0, NULL, 0)) {
    fprintf(stderr, fmtunexpected, prgname, *lnum);
    return false;
  }

  tTriangle raw = {{0,0,0}};
  for (size_t i = 0; 3 != i; ++i) {
    float x, y, z;  x = y = z = 0;
    memset(line, 0, LNBUFSZ);
    *lnum += 1;
    if (!getLine(line)) {
      fprintf(stderr, fmtunexpected, prgname, *lnum);
      return false;
    }
    else if (regexec(&rgx[reVertex], line, 0, NULL, 0)) {
      fprintf(stderr, fmtunexpected, prgname, *lnum);
      return false;
    }
    else if (3 != sscanf(line, " vertex %f %f %f", &x, &y, &z)) {
      fprintf(stderr, fmtunexpected, prgname, *lnum);
      return false;
    }
    raw[i].x = x, raw[i].y = y, raw[i].z = z;
  }

  bool const flag = addTriangle(raw);
  if (!flag) return false;

  memset(line, 0, LNBUFSZ);
  *lnum += 1;
  if (!getLine(line)) {
    fprintf(stderr, fmtunexpected, prgname, *lnum);
    return false;
  }
  if (regexec(&rgx[reLoopEnd], line, 0, NULL, 0)) {
    fprintf(stderr, fmtunexpected, prgname, *lnum);
    return false;
  }

  return true;
}

static bool procAscii(void) {
  char line[LNBUFSZ];
  uint32_t lnum = 1;  /* "solid " */

  while (true) {
    memset(line, 0, LNBUFSZ);
    lnum += 1;
    if (!getLine(line)) {
      fprintf(stderr, fmtunexpected, prgname, lnum);
      return false;
    }
    else if (!strncmp(ascends, line, strlen(ascends)))
      break;
    else if (regexec(&rgx[reFacet], line, 0, NULL, 0)) {
      fprintf(stderr, fmtunexpected, prgname, lnum);
      return false;
    }
    else if (!procOloop(&lnum))
      return false;

    memset(line, 0, LNBUFSZ);
    lnum += 1;
    if (!getLine(line)) {
      fprintf(stderr, fmtunexpected, prgname, lnum);
      return false;
    }
    else if (regexec(&rgx[reFacetEnd], line, 0, NULL, 0)) {
      fprintf(stderr, fmtunexpected, prgname, lnum);
      return false;
    }
  }

  return true;
}

static bool procBinary(void) {
  tXfer u;
  for (size_t i = 0; (size_t const)tcount != i; ++i) {
    tTriangle raw = {{0,0,0}};
    if (TSZ != fread(u.raw, sizeof(uint8_t), TSZ, stdin)) {
      if (feof(stdin))
        break;
      fprintf(stderr, "%s, oops reading triangle #%zu.\n", prgname, i);
      return false;
    }
    raw[0].x = u.val[3], raw[0].y = u.val[4], raw[0].z = u.val[5];
    raw[1].x = u.val[6], raw[1].y = u.val[7], raw[1].z = u.val[8];
    raw[2].x = u.val[9], raw[2].y = u.val[10], raw[2].z = u.val[11];
    bool const flag = addTriangle(raw);
    if (!flag) return false;
  }

  return true;
}

/* -------------------------------------------------------------------------- *
 * process header.  fixed size for binary format, first line when ascii.
 */

static bool procHeader(void) {
  char buf[LNBUFSZ] = {CHRNUL};
  size_t const n = strlen(ascident);
  if (n != fread(buf, sizeof(char), n, stdin)) {
    if (feof(stdin))
      return false;
    EEXIT;
  }

  if (strcmp(buf, ascident)) {  /* binary */
    size_t const m = 80 - n;
    if (m != fread(buf + n, sizeof(char), m, stdin)) {
      if (feof(stdin))
        return false;
      EEXIT;
    }
    else if (1 != fread(&tcount, sizeof(uint32_t), 1, stdin)) {
      if (feof(stdin))
        return false;
      EEXIT;
    }
    ascfmt = false;
  }
  else {  /* ascii */
    if (!fgets(buf + n, (int)(LNBUFSZ - 1 - n), stdin)) {
      if (feof(stdin))
        return false;
      EEXIT;
    }
    buf[strlen(buf) - 1] = CHRNUL;
  }

  if (diag) {
    fprintf(stderr, "[header] %s\n", buf);
    if (!ascfmt)
      fprintf(stderr, "[header] number triangles: %u\n", tcount);
  }

  return true;
}

/* ------------------------------------------------------------------------- */

static void tidyup(void) {
  gts_finalize();
  for (size_t i = 0; reEndRE != i; ++i)
    regfree(&rgx[i]);
  return;
}

static void setDefaults(void) {
  if (atexit(tidyup))
    EEXIT;
  else if (setreuid(getuid(), geteuid()))
    EEXIT;
  else if (!setlocale(LC_ALL, "POSIX"))
    EEXIT;
  for (size_t i = 0; reEndRE != i; ++i)
    if (regcomp(&rgx[i], repatterns[i], REG_EXTENDED|REG_NOSUB))
      EEXIT;
  S = gts_surface_new(gts_surface_class(), gts_face_class(),
          gts_edge_class(), gts_vertex_class());
  ascfmt = true;
  diag = false;
  return;
}

int main(int argc, char *argv[]) {
  setDefaults();

  if (1 < argc) {
    int o;
    opterr = 0;
    while (-1 != (o = getopt(argc, argv, prgopts)))
      switch (o) {
        case 'D': diag = true; break;
        case 'h': usageHelp(0);
        case 'V': usageVersion();
        default:  usageHelp(2);
      }
    if (argc != optind) {
      fprintf(stderr, "%s: oops, extra argument(s) on command line.\n",
              prgname);
      exit(EXIT_FAILURE);
    }
  }

  bool flag = procHeader();
  if (!flag) {
    fprintf(stderr, "%s: oops, data does not look like STL.\n", prgname);
    exit(EXIT_FAILURE);
  }

  flag = (ascfmt ? procAscii : procBinary)();
  if (flag) {
    if (diag)
      gts_surface_print_stats(S, stderr);
    gts_surface_write(S, stdout);
  }

  return flag ? EXIT_SUCCESS : EXIT_FAILURE;
}

/* ------------------------------------------------------------- *
 * the content above is covered by the GPLv3+, see file COPYING. *
 * copyright (c) 2020 jr <creature.eternal@gmail.com>.           *
 * all rights reserved.                                          *
 * ------------------------------------------------------------- *
 */
